#!/usr/bin/env perl

######################################################################
# blikII is a continuation of blik for FICS                          #
# Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>               #
#                                                                    #
# This file is part of blikII.                                       #
#                                                                    #
# blikII is free software: you can redistribute it and/or modify it  #
# under the terms of the GNU Affero General Public License as        #
# published by the Free Software Foundation, either version 3 of the #
# License, or (at your option) any later version.                    #
#                                                                    #
# blikII is distributed in the hope that it will be useful, but      #
# WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  #
# Affero General Public License for more details.                    #
#                                                                    #
# You should have received a copy of the GNU Affero General Public   #
# License along with blikII.  If not, see                            #
# <https://www.gnu.org/licenses/>.                                   #
######################################################################

use strict;
use warnings;
use Net::Telnet;
use Time::HiRes qw(time sleep);
use Getopt::Long qw(GetOptionsFromArray :config permute bundling);
use File::HomeDir ();
use v5.16;		# For fc
use subs qw(tell log);	# Override CORE functions with Main functions

my $name = 'blikII';
my $password;
my $version = '0.0';
my $operator = 'AsDaGo';

my $host = 'freechess.org';
my $port = 5000;

my $prompt = 'fics% ';

# A pattern that will match a handle on the server.
my $handle_pat = qr/[[:alpha:]]+/;

# The maximum length of a 'tell'.
my $max_tell = 200;

# The time (in seconds) that tells will be queued by the server.
my $tell_queue_time = 2;

# The time (in seconds) that we need to do something in order to
# prevent the server from kicking us off.
my $ping_time = 3600;

# The time of the last tell.
my $last_tell_time;

# The channels we will listen to.
my @channels = (120);

# Variables and their default values.
my @default_vars = (log => 1);

# Change @default_vars to the keys only, in the order they were
# given. %default_vars will be the hash.
my %default_vars = @default_vars;
for (my $i = 1; $i < @default_vars; $i++) {
    splice @default_vars, $i, 1;
}

# An environment variable used when we are restarted by a user. This
# is either the handle of a user or the number of a channel or the
# empty string for a shout.
my $restart_target_var = 'BLIKII_RESTART_TARGET';

# Where we will store and read configuration files.
my $config_dir;
{
    my $home = File::HomeDir->my_home;
    unless (defined $home) {
	warn 'No home directory; using current directory';
	$home = '.';
    }
    $config_dir = "$home/.blikii";
}
my $vars_dir = "$config_dir/vars";

my $usage = <<EOF;
Usage: $0 [OPTION]... [HOST [PORT]]
Connect to the chess server at HOST (default $host) through port PORT
(default $port) and start running.

  -n, --name=NAME         use NAME to log in rather than $name
  -v, --verbose           print verbose output
  -l, --log-file=FILE     log user communications to FILE
      --debug-file=FILE   log all I/O to FILE (NOTE: this does not
                          respect users' 'log' settings)
  -h, --help              print this help and exit
  -V, --version           print version information and exit
EOF

# Remember how we were run in case we need to restart ourself.
my @cmdline = ($0, @ARGV);

GetOptions
    'n|name=s'		=> \$name,
    'l|log-file=s'	=> \my $log_filename,
    'debug-file=s'	=> \my $debug_filename,
    'v|verbose'		=> \my $verbose,
    'h|help'		=> sub { print $usage; exit },
    'V|version'		=> sub { print "$name $version\n"; exit },
    or die $usage;

$host = shift if @ARGV;
$port = shift if @ARGV;
die $usage if @ARGV;

sub create_dir {
    foreach (@_) {
	unless (-d) {
	    die "$_ exists but is not a directory" if -e;
	    mkdir or die "Cannot create $_: $!";
	}
    }
}

# Create the configuration directory and variables directory if they
# don't already exist.
create_dir $config_dir, $vars_dir;

# The variables for each user.
my %user_vars;

# Load the user variables.
while (<$vars_dir/*>) {
    unless (m|^\Q$vars_dir/\E($handle_pat)$|) {
	warn "Invalid user variable file: $_";
	next;
    }

    my $filename = $_;
    my $handle = fc $1;
    my $file;
    unless (open $file, '<', $filename) {
	warn "Cannot open $filename: $!";
	next;
    }

    my %vars = %default_vars;
    while (<$file>) {
	chomp;
	unless (/^\s*(.+?)\s*=\s*(.+?)\s*$/) {
	    warn "$filename:$.: invalid line: $_\n";
	    next;
	}
	my ($key, $value) = ($1, $2);
	unless ($key =~ /^[[:alpha:]]+$/) {
	    warn "$filename:$.: invalid variable name: $key\n";
	    next;
	}
	unless ($value eq '0' || $value eq '1') {
	    warn "$filename:$.: invalid variable value: $value\n";
	    next;
	}
	$vars{$key} = $value + 0;
    }

    close $file or warn "Cannot close $filename: $_";

    $user_vars{$handle} = \%vars;
}

# Check if the users given by the first and second arguments are, in
# fact, the same users.
sub user_eq {
    fc $_[0] eq fc $_[1];
}

# Return variables for the user specified by the first argument (or $_
# if omitted). If more than one argument is given, return a list of
# the values of variables as specified by each of the arguments after
# the first argument.
sub get_vars {
    my $user = fc (@_ ? shift : $_);
    my %vars = exists $user_vars{$user} ?
	%{$user_vars{$user}} : %default_vars;
    return %vars unless @_;
    return map {
	die "Invalid variable: $_" unless exists $vars{$_};
	$vars{$_};
    } @_;
}

# Set the variables for the user specified by the first argument to
# the variables and values specified by the rest. Then, save the
# variable file. Returns true on success or false on error.
sub set_vars {
    die 'No args given' unless @_;
    my $user = fc shift;
    my %new_vars = @_;
    my %vars = exists $user_vars{$user} ?
	%{$user_vars{$user}} : %default_vars;

    foreach (keys %new_vars) {
	die "Invalid variable: $_" unless exists $vars{$_};
	$vars{$_} = $new_vars{$_};
    }
    $user_vars{$user} = \%vars;

    # Now save the variables.
    my $filename;
    while (<$vars_dir/*>) {
	m|^\Q$vars_dir/\E(.+)$| or die "Invalid file: $_";
	if (user_eq $user, $1) {
	    $filename = $_;
	    last;
	}
    }
    $filename = "$vars_dir/$user" unless defined $filename;

    my $file;
    unless (open $file, '>', $filename) {
	warn "Cannot open $filename: $_";
	return 0;
    }

    foreach (@default_vars) {
	unless (print $file "$_=$vars{$_}\n") {
	    warn "Cannot write to $filename: $_";
	    close $file;
	    return 0;
	}
    }

    unless (close $file) {
	warn "Cannot close $filename: $_";
	return 0;
    }

    return 1;
}

my $log_file;
if (defined $log_filename && !open $log_file, '>', $log_filename) {
    warn "Cannot open $log_filename: $!";
    undef $log_file;
    undef $log_filename;
}

my $debug_file;
if (defined $debug_filename &&
    ! open $debug_file, '>', $debug_filename) {
    warn "Cannot open $debug_filename: $!";
    undef $debug_file;
    undef $debug_filename;
}

# Used to check if we should log communications from a user.
my $current_user;

# Log the given arguments to standard output if $verbose, and
# $log_file if it is defined.
sub log {
    if (defined $current_user) {
	my ($log) = get_vars $current_user, 'log';
	return 0 unless $log;
	unshift @_, "$current_user: ";
    }
    my $ret = 1;
    $ret = (print @_) && $ret if $verbose;
    $ret = (print $log_file @_) && $ret if defined $log_file;
    return $ret;
}

my $telnet;

# Die unless the error was a timeout.
sub die_unless_timeout {
    die $_[0] unless $telnet->timed_out;
}

# Get a line and remove the "\r" at the beginning (if
# present). Returns the line on success, or undef on error.
sub getline {
    my $line = $telnet->getline(@_);
    return undef unless defined $line;
    chomp $line;
    $line =~ s/^\r//;
    return $line;
}

# Run a command through $telnet and return the output. If $verbose is
# true, also print the command.
sub cmd {
    my $cmd = join ' ', @_;
    log "> $cmd\n";

    # $telnet->cmd doesn't get the output correctly for some reason,
    # so we have to do it manually.
    return unless $telnet->print($cmd);
    my $first = getline Errmode => \&die_unless_timeout;
    return unless defined $first;
    $first =~ s/^$prompt//;
    my ($prematch) =
	$telnet->waitfor(Match => '/\r'.$prompt.'$/',
			 Errmode => \&die_unless_timeout);
    return unless defined $prematch;
    my @output = ($first, (split "\n", $prematch));
    foreach (@output) {
	s/^\r//;
	log "< $_\n";
    }
    return wantarray ? @output ? @output : '' : 1;
}

# Perform a tell while making sure it won't be queued. If the first
# argument is undef, shout instead of telling.
sub do_tell {
    my $user = shift;
    my $msg = @_ ? "@_" : $_;

    # Make sure our tell won't be queued.
    my $time = time;
    my $queued_until = $last_tell_time + $tell_queue_time;
    sleep $queued_until - $time if $queued_until > $time;

    my $ret = cmd (defined $user ? "tell $user $_" : "shout $_");
    $last_tell_time = time;
    return $ret;
}

# Tell the user specified by the first argument the message
# specified by the second argument (or $_ if not given). Returns
# true on success or false on failure.
sub tell_raw {
    my $user = shift;
    my $msg = @_ ? "@_" : $_;

    $msg =~ s/\n+$//;

    return 1 if $msg =~ /^\s*$/;

    # Split it into several lines if needed.
    my @msg = ($msg);
    while (1) {
	my $str = $msg[$#msg];
	my $len = length $str;
	last if $len <= $max_tell;
	my $i;
	for ($i = $max_tell; $i >= 0; $i--) {
	    last if (substr $str, $i, 1) eq ' ';
	}
	$i = $max_tell if $i < 0;
	my $start = substr $str, 0, $i;
	my $end = substr $str, $i + 1;
	$msg[$#msg] = $start;
	push @msg, $end;
    }

    # Perform the tell.
    my $ret = 1;
    foreach (@msg) {
	return 0 unless do_tell $user;
    }
    return 1;
}

# Call tell_raw after sanitizing the message.
sub tell {
    my $user = shift;
    my $msg = @_ ? "@_" : $_;

    # Sanitize the message.
    $msg =~ s/\s+/ /g;
    $msg =~ s/^\s+//;
    $msg =~ s/\s+$//;

    tell_raw $user, $msg;
}

# Get the variables of the user specified by the first argument (or
# $_) and return them as a hash. Returns an empty list on error.
sub get_server_vars {
    my $user = @_ ? $_[0] : $_;

    my @output = cmd "vars $user";
    return unless @output;

    if ($output[0] =~ /^Variable settings of (.+):$/ &&
	user_eq $1, $user) {
	shift @output;
    }
    else {
	warn "Got invalid variables for $user: $output[0]";
	return;
    }

    my %vars;
    foreach (@output) {
	s/^\r//;
	if (/^(\S+?): (.*)$/) {
	    $vars{$1} = $2;
	    next;
	}

	my @vars = split /\s+/;
	shift @vars if @vars && $vars[0] eq '';
	foreach (@vars) {
	    next unless /^(.+?)=(.*)$/;
	    $vars{$1} = $2;
	}
    }

    return %vars;
}

# Perform some actions before we quit.
sub before_quit {
    # Sleep a bit to allow the last tells to go through.
    sleep $tell_queue_time;
    $telnet->close;
    if (defined $log_file) {
	close $log_file or warn "Cannot close $log_filename: $!";
    }
    if (defined $debug_file) {
	close $debug_file or warn "Cannot close $debug_filename: $!";
    }
}

# Cause the bot to quit. Arguments are as for 'exit'.
sub quit {
    before_quit;
    print "Exiting\n" if $verbose;
    exit @_;
}

# Cause the bot to restart.
sub restart {
    before_quit;
    print "Restarting (executing '@cmdline')\n" if $verbose;
    exec $0 @cmdline;
    die "Cannot execute '@cmdline': $!";
}

# Superusers which can run special commands available only to them.
my @superusers = ($operator);

# Check if a given user is allowed to use superuser commands.
sub is_superuser {
    my $user = @_ ? $_[0] : $_;
    foreach (@superusers) {
	return 1 if user_eq $user, $_;
    }
    return 0;
}

# Check if a given user is the operator.
sub is_operator {
    my $user = @_ ? $_[0] : $_;
    user_eq $user, $operator;
}

my %commands;

# Tell the user specified by the first argument the usage for the
# command specified by the second argument.
sub tell_usage {
    my ($user, $cmd) = @_;
    my $usage = $commands{$cmd}->[1];
    tell $user, ("Usage: $cmd" . ($usage ? " $usage" : ''));
}

# A wrapper for GetOptionsFromArray so that error messages will be
# printed to the user who ran the command.
sub getopts {
    my $target = shift;
    local $SIG{__WARN__} = sub { tell $target, $_[0] };
    GetOptionsFromArray @_;
}

# Parse arguments for callbacks. Returns a list containing the user
# the tell came from, the channel the user used to send the tell (or
# undef for direct tells), and either the channel the tell came from
# or the user if it was a direct tell, and finally the name of the
# command and the arguments given to it.
sub parse_args {
    my ($user, $channel, @rest) = @_;
    my $target;
    if (defined $channel) {
	if ($channel eq 'shout' || $channel eq 'it') {
	    $target = undef;
	}
	else {
	    $target = $channel;
	}
    }
    else {
	$target = $user;
    }
    return ($user, $channel, $target, @rest);
}

# The commands available to users. Each value of the %commands hash is
# an array reference, the first item of which is the command callback,
# the second item of which is a short string describing the usage of
# that command, and the third item of which is a longer help
# string. The fourth item of the array reference is 0 if anyone can
# run the command, 1 if superuser permission is needed, or 2 if only
# the operator can run the command. The callbacks are called with the
# first argument as the user that ran the command, the second argument
# as the name the command was called as, and the rest of the arguments
# as the arguments given to the command. The callbacks should return
# true on success, or false for invalid syntax.
%commands = (
    quit => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args > 0;
	    tell $target, 'Shutting down...';
	    tell $operator, "$user shut me down.";
	    quit;
	},
	'', "Make $name exit the server.", 1
    ],
    restart => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args > 0;
	    tell $target, 'Restarting...';
	    tell $operator, "$user restarted me.";
	    $ENV{$restart_target_var} =
		(defined $target) ? $target : '';
	    restart;
	}, '', "Restart $name.", 1
    ],
    advertise => [
    	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
    	    push @args, 50 unless @args;
    	    foreach my $dest (@args) {
    		tell $dest, "Check it out! $operator whipped me up " .
    		    "just for fun. Try 'tell $name help', " .
    		    "'tell $name fortune', or " .
    		    "'tell $name figlet <MESSAGE>'."
    	    }
    	    return 1;
    	}, '[DESTINATION]...', 'Advertise myself to each ' .
    	'DESTINATION (user or channel, default 50).', 1
    ],
    suadd => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args < 1;
	    foreach (@args) {
		if (is_superuser) {
		    tell $target, "$_ is already a superuser.";
		}
		else {
		    push @superusers, $_;
		    tell $target, "$_ added to the superuser list.";
		    tell $_, 'You have been added to the ' .
			"superuser list by $user. Please don't " .
			'abuse your new powers.';
		}
	    }
	    return 1;
	}, 'USER...', 'Temporarily add USERs to the superuser list.', 2
    ],
    sudel => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args < 1;
	    foreach (@args) {
		my $i;
		for ($i = 0; $i < @superusers; $i++) {
		    last if user_eq $user, $superusers[$i];
		}
		if ($i >= @superusers) {
		    tell $target, "$_ is not a superuser.";
		}
		elsif (is_operator $user) {
		    tell $target, 'Cannot remove the operator ' .
			(@args > 1 ? "($operator) " : '') .
			'from the superuser list.';
		}
		else {
		    splice @superusers, $i, 1;
		    tell $target, "$_ removed from the superuser list.";
		    tell $_, 'You have been removed from the ' .
			"superuser list by $user.";
		}
	    }
	    return 1;
	}, 'USER...',
	'Temporarily remove USERs from the superuser list.', 2
    ],
    set => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 unless @args == 2;
	    my ($key, $value) = @args;

	    unless (exists $default_vars{$key}) {
		tell $target, "Invalid variable: '$key'";
		return 1;
	    }

	    if ($value =~ /^on|yes|true$/i) {
		$value = 1;
	    }
	    elsif ($value =~ /^off|no|false$/i) {
		$value = 0;
	    }
	    else {
		my $v = eval {
		    delete local $SIG{__DIE__};
		    local $SIG{__WARN__} = sub { die };
		    ($value + 0) ? 1 : 0;
		};
		unless (defined $v) {
		    tell $target, "Invalid value for '$key': $value";
		    return 1;
		}
		$value = $v;
	    }

	    tell $target, "Set $key to $value";
	    set_vars $user, $key => $value or tell $target,
		"Error saving variables; please contact $operator";

	    return 1;
	}, 'VARIABLE VALUE', 'Set VARIABLE to VALUE.', 0
    ],
    vars => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    my $vars_user = @args ? shift @args : $user;
	    return 0 if @args;

	    my %vars = get_vars $vars_user;
	    tell $target, "Variable settings of $vars_user:",
		map { "$_=$vars{$_}" } @default_vars;
	    return 1;
	}, '[USER]',
	'Print the variables for USER or yourself if omitted.', 0
    ],
    figlet => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    my $font = 'standard';
	    return 0 unless getopts $target, \@args,
		'f|font=s' => \$font;
	    return 0 unless @args;

	    if ($font =~ m|/|) {
		tell $target, 'Font must not be a full path';
		return 1;
	    }

	    my %vars = get_server_vars $user;
	    my $width;
	    if (exists $vars{width}) {
		$width = $vars{width};
	    }
	    else {
		warn "Cannot get width for $user";
		$width = 80;
	    }

	    $width -= length "$name(U) tells you: ";
	    $width = 1 if $width <= 0;

	    my $figlet;
	    unless (open $figlet, '-|', 'figlet',
		    '-w', $width, '-f', $font, '--', @args) {
		tell $target, 'Cannot run the figlet command. ' .
		    "Please contact $operator.";
		return 1;
	    }
	    my @message = <$figlet>;
	    close $figlet or tell $target, "Invalid font: '$font'";

	    foreach (@message) {
		chomp;
		if (/[^[:print:]]/) {
		    tell $target,
			"The font you chose ($font) cannot be " .
			'transferred. Please try a different font.';
		    return 1;
		}
	    }

	    tell_raw $target foreach @message;
	    return 1;
	}, '[-f FONT] MESSAGE', 'Print a message in ASCII art.', 0
    ],
    superusers => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args;
	    tell $target, "Superusers: @superusers";
	    return 1;
	}, '', 'List all superusers.', 0
    ],
    help => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args > 1;

	    if (@args) {
		my $cmd = $args[0];
		if (exists $commands{$cmd}) {
		    my ($help, $restricted) =
			@{$commands{$cmd}}[2, 3];
		    tell $target, 'RESTRICTED TO ' .
			(($restricted == 1) ? 'SUPERUSERS' :
			 'OPERATORS')
			if $restricted;
		    tell_usage $target, $cmd;
		    tell $target, $help;
		}
		else {
		    tell $target, "No help for $args[0].";
		}
	    }
	    else {
		# It would be cleaner to do
		#     \my (@cmds, @sucmds, @opcmds) = &get_commands;
		# But that feature is experimental, so we'd better
		# not...
		my ($cmds, $sucmds, $opcmds) = &get_commands;
		my @cmds = @$cmds;
		my @sucmds = @$sucmds;
		my @opcmds = @$opcmds;
		tell $target, "Available commands: @cmds";
		tell $target, "Superuser commands: @sucmds"
		    if is_superuser $user;
		tell $target, "Operator commands: @opcmds"
		    if is_operator $user;
		tell $target, "Try 'help <COMMAND>' for more info.";
	    }

	    return 1;
	},
	'[COMMAND]', 'Print the help for COMMAND.', 0
    ]
);

# Get all commands organized into regular, superuser, and operator
# commands, and sort them.
sub get_commands {
    my @cmds;
    my @sucmds;
    my @opcmds;
    foreach (keys %commands) {
	my $restrict = $commands{$_}->[3];
	if ($restrict == 0) {
	    push @cmds, $_;
	}
	elsif ($restrict == 1) {
	    push @sucmds, $_;
	}
	elsif ($restrict == 2) {
	    push @opcmds, $_;
	}
	else {
	    die "Invalid restriction value: $restrict";
	}
    }
    my @ret = (\@cmds, \@sucmds, \@opcmds);
    @$_ = sort @$_ foreach @ret;
    return @ret;
}

# These are like commands, except that they match a pattern rather
# than a specific string. The string to match against is always
# stripped of punctuation, and spaces are normalized (i.e.,
# s/\s+/ /). @phrases is an array rather than a hash to preserve the
# order of the keys, since the order in which patterns are tried
# matters. Each key is a pattern and each value is a callback. In
# the callbacks, $1, $2, etc. are available as groups in the
# pattern.
my @phrases = (
    qr/\b(?:$name)\b/i => sub {
	my ($user, $channel, $target) = parse_args @_;

	my $fortune = `fortune`;
	if (! defined $fortune || $?) {
	    tell $target, 'Cannot run the fortune command. ' .
		"Please contact $operator.";
	    return 1;
	}

	tell $target, $fortune;
    }
);

# Verify the integrity of the %commands hash.
foreach (values %commands) {
    die '%commands lacks integrity'
	unless @$_ == 4 && $_->[3] >= 0 && $_->[3] <= 2;
}

# Run a user command.
sub do_command {
    my ($user, $channel, $target, @cmd) = parse_args @_;

    # Only run commands if they were sent by a direct tell.
    return 0 if defined $channel;

    if (exists $commands{$cmd[0]}) {
	my ($func, $restricted) = @{$commands{$cmd[0]}}[0, 3];
	if ($restricted && ! is_superuser $user) {
	    tell $target, "You don't have permission to use " .
		"the '$cmd[0]' command.";
	}
	elsif ($restricted == 2 && ! is_operator $user) {
	    tell $target, "Only the operator ($operator) can " .
		"use the '$cmd[0]' command."
	}
	else {
	    $func->($user, $channel, @cmd)
		or tell_usage $target, $cmd[0];
	}
	return 1;
    }
    else {
	return 0;
    }
}

# Parse a phrase.
sub do_phrase {
    my $line = $_[2];
    $line =~ s/^\s+//;
    $line =~ s/\s+$//;
    $line =~ s/[^[:alpha:]]/ /;
    $line =~ s/\s+/ /;

    for (my $i = 0; $i < @phrases; $i += 2) {
	my ($pat, $func) = @phrases[$i, $i + 1];
	if ($line =~ $pat) {
	    $func->(@_);
	    return 1;
	}
    }
    return 0;
}

# Parse a user command or phrase.
sub parse_line {
    my $line;

    # Get the initial part of the command.
    $line = join ' ', @_;

    # Parse continuation lines.
    while (defined (getline Timeout => 0, Errmode => 'return')) {
	if (/^\\\s*(.*)$/) {
	    $line .= ' ' if $line;
	    $line .= $1;
	}
	else {
	    last;
	}
    }

    return ($line, (split /\s+/, $line));
}

# Connect to the server.
log "Opening connection to $host on port $port\n";
$telnet = new Net::Telnet
    Host	=> $host,
    Port	=> $port;
$telnet->input_log($debug_file) if defined $debug_file;
$telnet->open;

# Log in.
log "Waiting for login prompt\n";
$telnet->waitfor('/\rlogin: $/');
log "Logging in\n";
$telnet->print($name);
if (defined $password) {
    $telnet->waitfor('/\rpassword: $/');
    $telnet->print($password);
}
else {
    $telnet->print();
}

# Initialize the last tell time.
$last_tell_time = time;

$_ = sub {
    tell $operator, "Shutting down because I received SIG$_[0].";
    quit;
} foreach @SIG{qw(INT PWR QUIT TERM)};
$SIG{__DIE__} = sub {
    tell $operator, $_[0];
    before_quit;
};
$SIG{__WARN__} = sub { tell $operator, $_[0] };
$SIG{HUP} = sub {
    tell $operator, "Restarting because I received SIG$_[0].";
    restart;
};

# Wait for the prompt.
$telnet->waitfor('/\r'.$prompt.'$/');

# Wait a bit for all the initialization text to finish coming.
sleep 1;

# Run the initialization commands.
cmd 'set open 0';
cmd 'set seek 0';
cmd 'set width 240';
cmd "set interface $name $version";

# Set the finger notes.
{
    my @commands = map { @$_ } get_commands;
    my @finger_notes = split /\n/, <<EOF;
This is $name $version.
I am operated by $operator. Contact him with any questions.
In addition to responding to my name, I also have some special commands:
@commands
Try 'tell $name help' for more info.
Channel 120 (the former blik channel) is the blikII channel.
NOTE: In order to improve the bot, your communications will be temporarily logged, and $operator may read them. If you want to opt out of this, 'tell $name set log 0'.
======================================================================
Thanks to algra and mattuc for helping during the creation of $name.
Thanks to marcelk for writing the original blik. May his memory live on in $name.
EOF
    my $i;
    for ($i = 0; $i < @finger_notes; $i++) {
	cmd 'set', $i + 1, $finger_notes[$i];
    }
    cmd 'set', ++$i while $i < 10;
}

{
    # Initialize the channel list.

    # First get rid of all the channels we're currently listening
    # to.
    my @lines = cmd '=channel';
    shift @lines;
    my @cur_channels;
    foreach (@lines) {
	s/^\s+//;
	push @cur_channels, split /\s+/;
    }
    cmd "-channel $_" foreach @cur_channels;

    # Now add the channels we want to listen to.
    cmd "+channel $_" foreach @channels;
}

# If we were restarted by a user, let him know we're back up.
if (exists $ENV{$restart_target_var}) {
    tell (($ENV{$restart_target_var} || undef), "I'm back up!");
    delete $ENV{$restart_target_var};
}

# Contact the operator to let him know we connected.
tell $operator, "I'm online, $operator!";

my $timeout = 10;

# Start the main command loop.
COMMAND: while (1) {
    # Check if we should do something so that the server doesn't kick
    # us off. Allow for an extra grace period of $timeout.
    if (defined $ping_time) {
	my $time = time;
	tell $name, 'ping'
	    if $time - $last_tell_time > $ping_time - $timeout;
    }

    $_ = getline Timeout => $timeout,
	Errmode => \&die_unless_timeout;
    next unless defined;

    my $user;
    my $channel;
    my $line;
    if (/^($handle_pat).* tells you:\s+(.*)$/) {
	$user = $1;
	$line = $2;
    }
    elsif (/^($handle_pat)\((\d+)\):\s+(.*)$/) {
	$user = $1;
	$channel = $2;
	$line = $3;
    }
    elsif (/^($handle_pat) shouts:\s+(.*)$/) {
	$user = $1;
	$channel = 'shout';
	$line = $2;
    }
    elsif (/^--> ($handle_pat)\s+(.*)$/) {
	$user = $1;
	$channel = 'it';
	$line = $2;
    }
    else {
	next;
    }

    $current_user = $user;

    if (defined $channel) {
	# Make sure the tell came from one of the channels we're
	# listening to.
	my $found = 0;
	foreach (@channels) {
	    if ($channel == $_) {
		$found = 1;
		last;
	    }
	}
	die "Tell from unknown channel: $channel" unless $found;
    }

    # Ignore commands from ROBOadmin, adminBOT, and ourself
    # (generated by, for example, 'tell $name advertise $name').
    foreach ('ROBOadmin', 'adminBOT', $name) {
	next COMMAND if user_eq $user, $_;
    }

    my @cmd;
    ($line, @cmd) = parse_line $line;

    log "Line recieved from $user" .
	(defined $channel ? " (from channel $channel)" : '') .
	": @cmd\n";

    unless ((do_command $user, $channel, @cmd) ||
	    (do_phrase $user, $channel, $line) ||
	    (defined $channel)) {
	# It wasn't a valid command or phrase and it was a direct
	# tell; tell the user that we didn't understand.
	tell $user, "I'm sorry, but I don't understand.";
    }

    undef $current_user;
}

die 'This should not be happening.';
